from datetime import timedelta

from pl_mocks_and_fakes import MockInUnitTests, MockReason

from pl_user_io.delay import delay


@MockInUnitTests(MockReason.UNINVESTIGATED)
def pause_to_read(text: str) -> None:
    """
    Pauses for a length of time proportional to the length of the text.

    You probably want to use `pause_and_read`, instead, which also prints the text.
    """
    chars_per_second = 30
    min_read_time = 1
    read_time = max(len(text) / chars_per_second, min_read_time)
    delay(timedelta(seconds=read_time))



import itertools
from collections.abc import Callable, Generator
from contextlib import contextmanager
from textwrap import dedent
from typing import Any, TypedDict

import pytest
from pl_mocks_and_fakes import Fake, fake, fake_for, mock_for

from pl_user_io.display import display
from pl_user_io.loading_spinner import loading_spinner
from pl_user_io.str_input import StrInputParam, StrInputResponse, str_input
from pl_user_io.wait_for_enter import WAIT_FOR_ENTER_TEXT, wait_for_enter

InputMappingCallback = Callable[[str], str | None]


class PrintKwargs(TypedDict, total=False):
    end: str | None


@fake()
class UserIOFake(Fake):
    def __init__(self) -> None:
        def _wait_for_enter_side_effect() -> None:
            self._add_to_output_str(f"{WAIT_FOR_ENTER_TEXT} ✓\n")

        def _str_input_side_effect(prompt: StrInputParam) -> StrInputResponse:
            self._add_to_output_str(prompt)

            for callback in reversed(self.input_mapping_callbacks):
                result = callback(self.output_str())
                if result is not None:
                    self.input_mapping_callbacks.remove(callback)
                    return result

            msg = f"No user input matching output: `{self.output_str()}`"
            raise Exception(msg)

        def _display_side_effect(
            *args: list[Any],
            end: str | None = "\n",
            flush: bool = False,  # noqa: ARG001
        ) -> None:
            if len(args) > 0:
                self._add_to_output_str(str(args[0]))
            self._add_to_output_str(str("\n" if end is None else end))

        @contextmanager
        def _loading_spinner_side_effect(event: str) -> Generator[None]:
            try:
                yield
                self._add_to_output_str(f"✓ {event}\n")
            except Exception:
                self._add_to_output_str(f"✗ {event}\n")
                raise

        self._output_str: str = ""
        self.input_mapping_callbacks: list[InputMappingCallback] = []

        mock_for(wait_for_enter).side_effect = _wait_for_enter_side_effect
        mock_for(str_input).side_effect = _str_input_side_effect
        mock_for(loading_spinner).side_effect = _loading_spinner_side_effect
        mock_for(display).side_effect = _display_side_effect

    def _add_to_output_str(self, s: str) -> None:
        self._output_str += s

    def output_str(self) -> str:
        return self._output_str

    def clear_output(self) -> None:
        self._output_str = ""


def user_io_fake() -> UserIOFake:
    return fake_for(UserIOFake)


def assert_displayed(expected: str) -> None:
    assert expected in user_io_fake().output_str(), dedent("""\
        Expected output was not displayed.
        Expected Output:
        ----------------
        {expected}
        ----------------
        Full Output:
        ----------------
        {full_output}
        ----------------
    """).format(expected=expected, full_output=user_io_fake().output_str())


def assert_displayed_exactly_once(expected: str) -> None:
    assert_displayed(expected)
    output = user_io_fake().output_str()
    expected_count = output.count(expected)
    assert expected_count == 1, (
        f"Expected `{expected}` to be displayed once but it was displayed {expected_count} times in the output: `{output}`"
    )


def assert_not_displayed(expected: str) -> None:
    assert expected not in user_io_fake().output_str()


def assert_displayed_in_order(*outputs: str) -> None:
    output = user_io_fake().output_str()
    for before, after in itertools.pairwise(outputs):
        try:
            before_index = output.index(before)
        except ValueError:
            pytest.fail(
                f"Expected the output `{before}` to appear before the output `{after}`, but `{before}` did not appear at all. Complete output: `{output}`"
            )
        try:
            after_index = output.index(after)
        except ValueError:
            pytest.fail(
                f"Expected the output `{before}` to appear before the output `{after}`, but `{after}` did not appear at all. Complete output: `{output}`"
            )
        assert before_index < after_index, (
            f"Expected the output `{before}` to appear before `{after}` in the output: `{output}`"
        )



import pytest
from pl_mocks_and_fakes import mock_for

from pl_user_io.open_url import open_url
from pl_user_io.testing.user_io_fake import (
    assert_displayed,
    assert_displayed_exactly_once,
    assert_not_displayed,
)
from pl_user_io.wait_for_enter import WAIT_FOR_ENTER_TEXT


def assert_asked_yes_no(question: str) -> None:
    """Assert that the yes/no question was asked to the user."""
    assert_displayed(f"{question} (y/n)")


def assert_asked_yes_no_exactly_once(question: str) -> None:
    assert_displayed_exactly_once(f"{question} (y/n)")


def assert_not_asked_yes_no(question: str) -> None:
    assert_not_displayed(f"{question} (y/n)")


def assert_task_performed(task_str: str, url_opened: str | None = None) -> None:
    assert_displayed(f"{task_str}\n{WAIT_FOR_ENTER_TEXT}")
    if url_opened is not None:
        urls_opened = [call[0][0] for call in mock_for(open_url).call_args_list]
        assert url_opened in urls_opened


def assert_task_not_performed(task_str: str, url_opened: str | None = None) -> None:
    with pytest.raises(AssertionError):
        assert_task_performed(task_str, url_opened=url_opened)


def assert_loading_spinner_displayed(event: str) -> None:
    assert_displayed(f"✓ {event}")



from datetime import timedelta
from time import sleep

from pl_mocks_and_fakes import MockInUnitTests, MockReason


@MockInUnitTests(MockReason.UNINVESTIGATED)
def delay(t: timedelta) -> None:
    """
    Pauses thread execution.

    This should be used instead of `time.sleep` because the behavior is documented better.

    Computer sleep interferes with this function, extending it by the amount of time the computer was asleep. For this reason it is recommended to pass small t values.
    """
    seconds = t.total_seconds()
    sleep(seconds)



import subprocess
from pathlib import Path

from pl_mocks_and_fakes import MockInUnitTests, MockReason

FIREFOX_ABSOLUTE_PATH = Path("/snap/bin/firefox")


@MockInUnitTests(MockReason.UNINVESTIGATED)
def open_url(url: str) -> None:
    # Using run_program hangs until you quit the browser, which is not what we want. We just want to open the URL and move on.
    subprocess.Popen(
        [FIREFOX_ABSOLUTE_PATH, url],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )



from collections.abc import Generator
from contextlib import contextmanager

from pl_mocks_and_fakes import MockInUnitTests, MockReason
from rich import print as rprint
from rich import status
from rich.style import Style


@MockInUnitTests(MockReason.UNMITIGATED_SIDE_EFFECT)
@contextmanager
def loading_spinner(event: str) -> Generator[None]:
    s = status.Status(event, spinner="dots", spinner_style=Style(color="cyan"))
    s.start()
    try:
        yield
        rprint(f"[green]✓[/green] {event}")
    except Exception:
        rprint(f"[red]✗[/red] {event}")
        raise
    finally:
        s.stop()



from pl_mocks_and_fakes import (
    HUMAN_INTERACTION_MOCK_REASONS,
    MockInUnitTests,
)

StrInputParam = str
StrInputResponse = str


@MockInUnitTests(*HUMAN_INTERACTION_MOCK_REASONS)
def str_input(prompt: StrInputParam) -> StrInputResponse:
    return input(prompt + " ")



from pl_user_io.display import display
from pl_user_io.str_input import str_input


def yes_no(question: str) -> bool:
    """Asks the user a yes/no question and returns their answer ("y" or "n") as a boolean."""
    response = str_input(f"{question} (y/n)")

    if response == "y":
        display()
        return True
    if response == "n":
        display()
        return False
    return yes_no(question)
